﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Media.Imaging;
using System.Collections.Generic;
using System.Linq;
using AnimationCreator.Model;
using System.Xml.Linq;

namespace AnimationCreator.Utilities
{
	public class TextureCreator
	{
		#region Properties

		/// <summary>
		/// The maximum width of the final texture. 
		/// Only used if the texture size shall not be a power of 2.
		/// </summary>
		public int MaxTextureWidth
		{
			get;
			set;
		}

		/// <summary>
		/// True if the texture size should be a power of two, otherwise false.
		/// </summary>
		public bool MakeTextureSizePowerOf2
		{
			get;
			set;
		}

		/// <summary>
		/// True if the texture should be a square, otherwise false.
		/// </summary>
		public bool MakeTextureSquare
		{
			get;
			set;
		}

		#endregion

		#region Constructor

		public TextureCreator()
		{
			// default values
			MaxTextureWidth = 1024;
			MakeTextureSizePowerOf2 = false;
			MakeTextureSquare = false;
		}

		#endregion

		#region Implementation

		/// <summary>
		/// Creates the final texture from the given source images, using the parameter settings of the class properties.
		/// </summary>
		/// <param name="sourceImages">The source images to use.</param>
		/// <returns>The final texture.</returns>
		public TextureResult CreateTexture(IEnumerable<SourceImage> sourceImages)
		{
			if (sourceImages == null)
			{
				return null;
			}

			int maxWidth = 0;
			int maxHeight = 0;
			int sourceImagesCount = sourceImages.Count();

			if (sourceImagesCount == 0)
			{
				return null;
			}

			// determine the max width and height
			foreach (var sourceImage in sourceImages)
			{
				var bitmap = sourceImage.Bitmap;

				if (bitmap.PixelWidth > maxWidth)
				{
					maxWidth = bitmap.PixelWidth;
				}

				if (bitmap.PixelHeight > maxHeight)
				{
					maxHeight = bitmap.PixelHeight;
				}
			}

			if (maxWidth < 1 || maxHeight < 1)
			{
				return null;
			}

			// take the maximum of width and height as basis
			int maxEdgeLength = Math.Max(maxWidth, maxHeight);
			
			// calculate the texture size
			int textureWidth = CalculateTextureWidth(sourceImagesCount, maxEdgeLength);
			int textureHeight = CalculateTextureHeight(sourceImagesCount, textureWidth, maxEdgeLength);

			// create and clear target bitmap
			WriteableBitmap bmp = new WriteableBitmap(textureWidth, textureHeight);
			bmp.Clear(Colors.Transparent);

			// copy source images
			XDocument description = CopySourceImages(sourceImages, bmp, textureWidth, maxEdgeLength);

			// create the result
			TextureResult result = new TextureResult();
			result.Texture = bmp;
			result.Description = description;

			return result;
		}

		private int CalculateTextureWidth(int sourceImagesCount, int maxEdgeLength)
		{
			// calculates the final texture width
			int textureWidth = 0;

			if (MakeTextureSizePowerOf2)
			{
				// calculate the required are and square root for a first shot
				int requiredArea = maxEdgeLength * maxEdgeLength * sourceImagesCount;
				textureWidth = (int)Math.Ceiling(Math.Sqrt(requiredArea));

				// get the fractional power of 2 and round it to the next one
				double asPowerOf2 = Math.Log(textureWidth, 2.0);
				int roundedPowerOf2 = (int)Math.Ceiling(asPowerOf2);

				// and there we have the final width
				textureWidth = (int)Math.Pow(2.0, roundedPowerOf2);
			}
			else
			{
				// is this even possible?
				if (MaxTextureWidth < maxEdgeLength)
				{
					throw new ArgumentException("Creating the texture is not possible: maximum width restriction is smaller than required minimum width (" + maxEdgeLength + ").");
				}

				textureWidth = Math.Min(MaxTextureWidth, maxEdgeLength * sourceImagesCount);

				// if the texture width was cropped to MaxTextureWidth, 
				// the final size might not be correct yet.
				int framesPerRow = textureWidth / maxEdgeLength; // integer division
				textureWidth = framesPerRow * maxEdgeLength; // now it is correct
			}

			return textureWidth;
		}

		private int CalculateTextureHeight(int sourceImagesCount, int textureWidth, int maxEdgeLength)
		{
			// => first calc the frames per row (integer division)
			int framesPerRow = textureWidth / maxEdgeLength;

			// now calc the rows
			int rows = (int)Math.Ceiling((double)sourceImagesCount / framesPerRow);

			// and finally the height
			int textureHeight = rows * maxEdgeLength;

			// correct the height if necessary
			if (MakeTextureSquare)
			{
				// check whether creating a square texture is possible at all
				if (textureHeight > textureWidth)
				{
					throw new ArgumentException("Creating a square texture is not possible with this width restriction.");
				}

				return textureWidth;
			}
			else
			{	
				// make it a power of 2?
				if (MakeTextureSizePowerOf2)
				{
					// get the fractional power of 2, then round it
					double asPowerOf2 = Math.Log(textureHeight, 2.0);
					int roundedPowerOf2 = (int)Math.Ceiling(asPowerOf2);

					// this is the final height
					textureHeight = (int)Math.Pow(2.0, roundedPowerOf2);
				}

				return textureHeight;
			}
		}

		private XDocument CopySourceImages(IEnumerable<SourceImage> sourceImages, WriteableBitmap bmp, int textureWidth, int maxEdgeLength)
		{
			int currentWidth = 0;
			int currentHeight = 0;

			// holds the description
			XDocument description = new XDocument();
			XElement animation = new XElement("Animation");
			description.Add(animation);

			// copy the source images to the target
			for (int i = 0; i < sourceImages.Count(); i++)
			{
				var bitmap = sourceImages.ElementAt(i).Bitmap;

				// new line?
				if (currentWidth + maxEdgeLength > textureWidth)
				{
					currentWidth = 0;
					currentHeight += maxEdgeLength;
				}

				// calc destination rectangle
				Rect destination = new Rect(currentWidth, currentHeight, bitmap.PixelWidth, bitmap.PixelHeight);
				Rect source = new Rect(0, 0, bitmap.PixelWidth, bitmap.PixelHeight);

				// copy!
				bmp.Blit(destination, bitmap, source);

				// document everything
				XElement frameElement = new XElement("Frame",
					new XAttribute("Id", i),
					new XAttribute("Name", System.IO.Path.GetFileNameWithoutExtension(sourceImages.ElementAt(i).Filename)),
					new XAttribute("X", currentWidth),
					new XAttribute("Y", currentHeight),
					new XAttribute("Width", bitmap.PixelWidth),
					new XAttribute("Height", bitmap.PixelHeight)
					);
				animation.Add(frameElement);

				// move to the next spot
				currentWidth += maxEdgeLength;
			}

			return description;
		}

		#endregion
	}
}
