﻿using GalaSoft.MvvmLight;
using System.Windows.Media.Imaging;
using System.Collections.ObjectModel;
using System.Windows.Input;
using GalaSoft.MvvmLight.Command;
using System.Windows.Controls;
using System.Linq;
using System.IO;
using AnimationCreator.Model;
using System.Windows.Data;
using System.ComponentModel;
using AnimationCreator.Utilities;
using ImageTools.IO.Png;
using ImageTools;
using System.Collections.Generic;
using System.Windows;
using System;

namespace AnimationCreator.ViewModel
{
	public class MainViewModel : ViewModelBase
	{
		private SourceImage _selectedSourceImage;
		private TextureResult _textureResult;
		private string _errors;

		#region Properties

		public string Errors
		{
			get
			{
				return _errors;
			}
			set
			{
				_errors = value;
				RaisePropertyChanged("Errors");
				RaisePropertyChanged("ErrorsVisibility");
			}
		}

		public Visibility ErrorsVisibility
		{
			get
			{
				if (string.IsNullOrWhiteSpace(Errors))
				{
					return Visibility.Collapsed;
				}
				else
				{
					return Visibility.Visible;
				}
			}
		}

		private TextureCreator TextureCreator
		{
			get;
			set;
		}

		public TextureResult TextureResult
		{
			get
			{
				return _textureResult;
			}
			set
			{
				_textureResult = value;

				// need to raise some changed events here
				RaisePropertyChanged("TextureResult");
				RaisePropertyChanged("TextureResultVisibility");
				RaisePropertyChanged("InvertedTextureResultVisibility");

				// re-evaluate commands
				(SaveTextureCommand as RelayCommand).RaiseCanExecuteChanged();
				(SaveDescriptionCommand as RelayCommand).RaiseCanExecuteChanged();
			}
		}

		public Visibility TextureResultVisibility
		{
			get
			{
				if (TextureResult == null)
				{
					return Visibility.Collapsed;
				}
				else
				{
					return Visibility.Visible;
				}
			}
		}

		public Visibility InvertedTextureResultVisibility
		{
			get
			{
				if (TextureResult == null)
				{
					return Visibility.Visible;
				}
				else
				{
					return Visibility.Collapsed;
				}
			}
		}

		public int MaxTextureWidth
		{
			get
			{
				return TextureCreator.MaxTextureWidth;
			}
			set
			{
				TextureCreator.MaxTextureWidth = value;

				// re-create the texture automatically
				CreateTexture();
			}
		}

		public bool MakeTextureSizePowerOf2
		{
			get
			{
				return TextureCreator.MakeTextureSizePowerOf2;
			}
			set
			{
				TextureCreator.MakeTextureSizePowerOf2 = value;

				// re-create the texture automatically
				CreateTexture();
			}
		}

		public bool MakeTextureSquare
		{
			get
			{
				return TextureCreator.MakeTextureSquare;
			}
			set
			{
				TextureCreator.MakeTextureSquare = value;

				// re-create the texture automatically
				CreateTexture();
			}
		}
		
		public ObservableCollection<SourceImage> SourceImages
		{
			get;
			private set;
		}

		public SourceImage SelectedSourceImage
		{
			get
			{
				return _selectedSourceImage;
			}
			set
			{
				_selectedSourceImage = value;
				RaisePropertyChanged("SelectedSourceImage");

				// re-evaluate commands
				(RemoveSelectedSourceImageCommand as RelayCommand).RaiseCanExecuteChanged();
				(MoveSelectedSourceImageUpCommand as RelayCommand).RaiseCanExecuteChanged();
				(MoveSelectedSourceImageDownCommand as RelayCommand).RaiseCanExecuteChanged();
			}
		}

		public ICommand SelectSourceImagesCommand
		{
			get;
			private set;
		}

		public ICommand ClearSourceImagesCommand
		{
			get;
			private set;
		}
		
		public ICommand RemoveSelectedSourceImageCommand
		{
			get;
			private set;
		}

		public ICommand MoveSelectedSourceImageUpCommand
		{
			get;
			private set;
		}

		public ICommand MoveSelectedSourceImageDownCommand
		{
			get;
			private set;
		}

		public ICommand SaveTextureCommand
		{
			get;
			private set;
		}

		public ICommand SaveDescriptionCommand
		{
			get;
			private set;
		}

		public ICommand InstallOfflineCommand
		{
			get;
			private set;
		}

		public Visibility InstallOfflineCommandVisibility
		{
			get
			{
				return App.Current.IsRunningOutOfBrowser ? Visibility.Collapsed : Visibility.Visible;
			}
		}

		#endregion

		#region Constructor and helper methods

		/// <summary>
		/// Initializes a new instance of the MainViewModel class.
		/// </summary>
		public MainViewModel()
		{
			TextureCreator = new TextureCreator();
			SourceImages = new ObservableCollection<SourceImage>();

			SelectSourceImagesCommand = new RelayCommand<IEnumerable<FileInfo>>(SelectSourceImages);
			ClearSourceImagesCommand = new RelayCommand(ClearSourceImages, () => SourceImages.Count > 0);
			RemoveSelectedSourceImageCommand = new RelayCommand(RemoveSelectedSourceImage, () => SelectedSourceImage != null);
			MoveSelectedSourceImageUpCommand = new RelayCommand(MoveSelectedSourceImageUp, () => SelectedSourceImage != null && SourceImages.IndexOf(SelectedSourceImage) > 0);
			MoveSelectedSourceImageDownCommand = new RelayCommand(MoveSelectedSourceImageDown, () => SelectedSourceImage != null && SourceImages.IndexOf(SelectedSourceImage) < SourceImages.Count - 1);
			SaveTextureCommand = new RelayCommand(SaveTexture, () => TextureResult != null && TextureResult.Texture != null);
			SaveDescriptionCommand = new RelayCommand(SaveDescription, () => TextureResult != null && TextureResult.Description != null);
			InstallOfflineCommand = new RelayCommand(InstallOffline, () => !App.Current.IsRunningOutOfBrowser);

			// Todo: create design-time data
			if (IsInDesignMode)
			{
				// Code runs in Blend --> create design time data.
			}
			else
			{
				// Code runs "for real"
			}
		}

		private void ClearErrors()
		{
			Errors = null;
		}

		private void AddError(string error)
		{
			if (Errors == null)
			{
				Errors = error;
			}
			else
			{
				Errors += Environment.NewLine;
				Errors += error;
			}
		}

		#endregion

		#region Command implementations

		private void SelectSourceImages(IEnumerable<FileInfo> fileInfos)
		{
			ClearErrors();

			// make sure we have some files selected
			if (fileInfos == null || fileInfos.Count() == 0)
			{
				AddError("Please select some files.");
				return;
			}

			// check file extensions
			foreach (var fileInfo in fileInfos)
			{
				if (fileInfo.Extension.ToUpper() != ".PNG")
				{
					AddError("Only PNG files are supported at the moment.");
					return;
				}
			}

			// first sort the files list
			var files = fileInfos.ToList();
			try
			{
				files.Sort(new System.Comparison<FileInfo>((a, b) =>
				{
					return a.Name.CompareTo(b.Name);
				}));
			}
			catch (Exception ex)
			{
				AddError("Error while sorting the source files: " + ex.Message);
				return;
			}

			try
			{
				// now do the work
				LoadSourceImages(files);
			}
			catch (Exception ex)
			{
				AddError("Error while loading the source files: " + ex.Message);
				return;
			}

			// re-create the texture automatically
			CreateTexture();

			// re-evaluate commands
			(ClearSourceImagesCommand as RelayCommand).RaiseCanExecuteChanged();
			(MoveSelectedSourceImageUpCommand as RelayCommand).RaiseCanExecuteChanged();
			(MoveSelectedSourceImageDownCommand as RelayCommand).RaiseCanExecuteChanged();
		}
		
		private void LoadSourceImages(IList<FileInfo> fileInfos)
		{
			foreach (FileInfo fileInfo in fileInfos)
			{
				using (FileStream fs = fileInfo.OpenRead())
				{
					// we load a bitmap image from each file and create a writeable bitmap from it
					BitmapImage image = new BitmapImage();
					image.SetSource(fs);
					WriteableBitmap bmp = new WriteableBitmap(image);

					// create entry 
					SourceImage sourceImage = new SourceImage(bmp, fileInfo.Name);
					SourceImages.Add(sourceImage);
				}
			}
		}

		private void ClearSourceImages()
		{
			ClearErrors();

			SourceImages.Clear();

			// re-create the texture automatically
			CreateTexture();

			// re-evaluate commands
			(ClearSourceImagesCommand as RelayCommand).RaiseCanExecuteChanged();
		}

		private void RemoveSelectedSourceImage()
		{
			ClearErrors();

			if (SelectedSourceImage != null)
			{
				SourceImages.Remove(SelectedSourceImage);

				// re-create the texture automatically
				CreateTexture();
			}
		}

		private void MoveSelectedSourceImageUp()
		{
			ClearErrors();

			// store a reference to the current image,
			// we need it later
			var selectedImage = SelectedSourceImage;

			if (selectedImage != null)
			{
				int index = SourceImages.IndexOf(selectedImage);

				// only do something if it's feasible
				if (index == 0)
				{
					return;
				}

				index--;

				// remove and re-insert
				SourceImages.Remove(selectedImage);
				SourceImages.Insert(index, selectedImage);

				// re-select
				SelectedSourceImage = selectedImage;

				// re-create the texture automatically
				CreateTexture();
			}
		}

		private void MoveSelectedSourceImageDown()
		{
			ClearErrors();

			// store a reference to the currently selected image, we need it below
			var selectedImage = SelectedSourceImage;

			if (selectedImage != null)
			{
				int index = SourceImages.IndexOf(selectedImage);

				// only do something if it's feasible
				if (index == SourceImages.Count - 1)
				{
					return;
				}

				index++;

				// remove and re-insert
				SourceImages.Remove(selectedImage);
				SourceImages.Insert(index, selectedImage);

				// re-select
				SelectedSourceImage = selectedImage;

				// re-create the texture automatically
				CreateTexture();
			}
		}

		private void CreateTexture()
		{
			ClearErrors();

			try
			{
				// let the creator do all the work
				TextureResult = TextureCreator.CreateTexture(SourceImages);
			}
			catch (Exception ex)
			{
				TextureResult = null;
				AddError("Error while creating the texture: " + ex.Message);				
				return;
			}
		}

		private void SaveTexture()
		{
			ClearErrors();
			
			try
			{
				SaveFileDialog dlg = new SaveFileDialog();
				dlg.DefaultExt = ".png";
				dlg.Filter = "PNG (*.png)|*.png";
				bool? result = dlg.ShowDialog();

				if (result.HasValue && result.Value)
				{
					using (Stream stream = dlg.OpenFile())
					{
						// create an image tools image first, then save as PNG
						ExtendedImage image = TextureResult.Texture.ToImage();
						PngEncoder encoder = new PngEncoder();
						encoder.Encode(image, stream);
					}
				}
			}
			catch (Exception ex)
			{
				AddError("Error while saving the texture: " + ex.Message);
				return;
			}
		}
		
		private void SaveDescription()
		{
			ClearErrors();

			try
			{
				SaveFileDialog dlg = new SaveFileDialog();
				dlg.DefaultExt = ".xml";
				dlg.Filter = "XML (*.xml)|*.xml";
				bool? result = dlg.ShowDialog();

				if (result.HasValue && result.Value)
				{
					using (Stream stream = dlg.OpenFile())
					{
						// simply let the XDocument save itself
						TextureResult.Description.Save(stream);
					}
				}
			}
			catch (Exception ex)
			{
				AddError("Error while saving the description file: " + ex.Message);
				return;
			}
		}

		private void InstallOffline()
		{
			if (!App.Current.IsRunningOutOfBrowser)
			{
				App.Current.Install();
			}
		}

		#endregion
	}
}